------------------------------------------------------------------
-- PROJECT:     HiCoVec Processor
--
-- ENTITY:      Memory Interface
--
-- PURPOSE:     Speicherinterface der CPU
-- 
-- 
-- Created by: Harald Manske
--
-- Prof. Gundolf Kiefer
-- Hochschule Augsburg
--
-- 
--
-- 
-- DATE CREATE: 
-- LAST CHANGE: 2008-11-24 FROM: Andreas Mueller
-- VERSION:     0.1
-----------------------------------------------------------------



library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

use work.cfg.all;
use work.datatypes.all;

entity hicovec is
	port(
-- Allgemeine Signale
		clk_hicovec:	in std_logic;
		reset:		in std_logic;
-- Debug Signale direkt vom Prozessor        
		dbg_a:		out std_logic_vector(31 downto 0);
		dbg_x:		out std_logic_vector(31 downto 0);
		dbg_y:		out std_logic_vector(31 downto 0);
		dbg_ir: 	out std_logic_vector(31 downto 0);
		dbg_ic: 	out std_logic_vector(31 downto 0);
		dbg_carry: 	out std_logic;
		dbg_zero: 	out std_logic;
		dbg_ir_ready: 	out std_logic;
		dbg_halted: 	out std_logic;
-- Signale vom "Memory Interface" zum VISCY-System
		ready:		in std_logic;
		wr:             out std_logic;
		rd:             out std_logic;
		address:        out std_logic_vector(31 downto 0);
		data:		inout  std_logic_vector(31 downto 0)
	);
end hicovec;

architecture rtl of hicovec is

	component cpu
		port(
			clk: 		in std_logic;
			reset: 		in std_logic;
			dbg_a: 		out std_logic_vector(31 downto 0);
			dbg_x: 		out std_logic_vector(31 downto 0);
			dbg_y: 		out std_logic_vector(31 downto 0);
			dbg_ir: 	out std_logic_vector(31 downto 0);
			dbg_ic: 	out std_logic_vector(31 downto 0);
			dbg_carry: 	out std_logic;
			dbg_zero: 	out std_logic;
			dbg_ir_ready: out std_logic;
			dbg_halted: out std_logic;
			mem_data_in: in std_logic_vector(31 downto 0);
			mem_data_out: out std_logic_vector(31 downto 0);
			mem_vdata_in: in vectordata_type;
			mem_vdata_out: out vectordata_type;
			mem_address: out std_logic_vector(31 downto 0);
			mem_access: out std_logic_vector(2 downto 0);
			mem_ready: 	in std_logic
		);
	end component;

	component memoryinterface
		port (
			clk:		in  std_logic;
			reset:		in  std_logic;
			address:	in  std_logic_vector(31 downto 0) := (others => '0'); 
			access_type:	in  std_logic_vector(2 downto 0) := "000";	-- we/wr, oe/rd, cs
			data_in:	in  std_logic_vector(31 downto 0);		-- Aus Skalareinheit
			vdata_in:	in  vectordata_type; 				-- Aus Vektoreinheit
			data_out:	out std_logic_vector(31 downto 0);		-- Zur Skalareinheit
			vdata_out:	out vectordata_type;				-- Zur Vektoreinheit
			cpu_ready:	out std_logic;					-- Ready zur CPU
			mem_ready:	in  std_logic;					-- Ready vom Speicher
			wr:		out std_logic;					-- we und en ersetzen
			rd:		out std_logic;					-- en ersetzen
			addr:		out std_logic_vector(31 downto 0);		-- Adresse an Ausgang (zum Memory)
			di:		out std_logic_vector(31 downto 0);		-- Dateneingang vom Memory
			do:		in std_logic_vector(31 downto 0)		-- Datenausgang zum Memory
		);
	end component;

	component db_multiplexer
		 port (   
			wr:				in std_logic;
			d_in:  			out std_logic_vector(31 downto 0);
			d_out:  		in std_logic_vector(31 downto 0);
			data:   		inout std_logic_vector(31 downto 0)
		);
	end component;		

	for cpu_impl:			cpu		use entity work.cpu(rtl);
	for memoryinterface_impl:	memoryinterface	use entity work.memoryinterface(rtl);
	for db_multiplexer_impl:	db_multiplexer	use entity work.db_multiplexer(rtl);
	

	signal s_reset:			std_logic;
-- SIGNALs zwischen CPU und Memory Interface
	signal s_mem_data_in: 		std_logic_vector(31 downto 0);	-- Datenleitung von Mem. Int. zur CPU
	signal s_mem_vdata_in: 		vectordata_type;		-- V-Datenleitung von Mem. Int. zur CPU
	signal s_mem_ready:		std_logic;			-- Ready Signal vom Mem. Int. zur CPU
	
	signal s_mem_data_out: 		std_logic_vector(31 downto 0);	-- Datenleitung von CPU zum Mem. Int. 
	signal s_mem_vdata_out: 	vectordata_type;		-- V-Datenleitung von CPU zum Mem. Int. 
	signal s_mem_address_out: 	std_logic_vector(31 downto 0);	-- Adressleitung von CPU zum Mem. Int. 
	signal s_mem_access:		std_logic_vector(2 downto 0);	-- Speicherzugriffsart (we, oe, cs)
	
	signal s_d_in:			std_logic_vector(31 downto 0);	-- Leitung zum db-muxer
	signal s_d_out:			std_logic_vector(31 downto 0);	-- Leitung zum db-muxer
	signal s_wr:			std_logic;			-- Write Enable Leitung
	

begin

	cpu_impl: cpu
		port map (
			clk 		=> clk_hicovec,
			reset 		=> s_reset,
			dbg_a 		=> dbg_a,	-- Debugg Signale
			dbg_x 		=> dbg_x,
			dbg_y 		=> dbg_y,
			dbg_ir 		=> dbg_ir,
			dbg_ic 		=> dbg_ic,
			dbg_carry 	=> dbg_carry,
			dbg_zero 	=> dbg_zero,
			dbg_ir_ready 	=> dbg_ir_ready,
			dbg_halted 	=> dbg_halted,
			mem_data_in 	=> s_mem_data_in,	-- Signale CPU <--> Memory Interface
			mem_data_out 	=> s_mem_data_out,
			mem_vdata_in 	=> s_mem_vdata_in,
			mem_vdata_out 	=> s_mem_vdata_out,
			mem_address 	=> s_mem_address_out,
			mem_access 	=> s_mem_access,
			mem_ready 	=> s_mem_ready
        );

	memoryinterface_impl: memoryinterface 
		port map (
			clk		=> clk_hicovec,
			reset		=> s_reset,
			address		=> s_mem_address_out,	-- Signale CPU <--> Memory Interface
			access_type	=> s_mem_access, 
			data_in		=> s_mem_data_out,
			vdata_in	=> s_mem_vdata_out,
			data_out	=> s_mem_data_in,
			vdata_out	=> s_mem_vdata_in,
			cpu_ready	=> s_mem_ready,		-- Ready zur CPU
			mem_ready	=> ready,
			wr		=> s_wr,
			rd		=> rd,
			addr		=> address,
			di		=> s_d_out,
			do		=> s_d_in
		);
		
	db_multiplexer_impl: db_multiplexer
		port map (
			wr		=> s_wr,
			d_in		=> s_d_in,
			d_out		=> s_d_out,
			data		=> data
		);
		
-- Glue logic
		wr 	<= s_wr;
		s_reset <= reset;


    
end;

