------------------------------------------------------------------
-- PROJECT:     Der Hicovec Prozessor in Aktion
--
-- ENTITY:      Memory Interface
--
-- PURPOSE:     Speicherinterface der CPU
-- 
-- System to
-- Connect,
-- Operate ,
-- Transfer data and
-- Compute with the
-- Hicovec 
--
-- TEAM:
--
-- Jakob Golus
-- David Lucinkiewicz
-- Andreas Mueller
-- Florian Richter	
-- Sebastian Rigorth
-- Soeren Ruehm
--
-- 
-- DATE: 2008-11-11
-- VERSION:     0.1
-----------------------------------------------------------------


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

use work.cfg.all;
use work.datatypes.all;

entity memoryinterface is
	port (   
	clk:		in  std_logic;
	reset:		in  std_logic;
	address:	in  std_logic_vector(31 downto 0) := (others => '0'); 
	access_type:	in  std_logic_vector(2 downto 0) := "000";	-- we/wr, oe/rd, cs
	data_in:	in  std_logic_vector(31 downto 0);		-- Aus Skalareinheit
	vdata_in:	in  vectordata_type; 				-- Aus Vektoreinheit
	data_out:	out std_logic_vector(31 downto 0);		-- Zur Skalareinheit
	vdata_out:	out vectordata_type;				-- Zur Vektoreinheit
	cpu_ready:	out std_logic;
	mem_ready:	in  std_logic;
	wr:		out std_logic;					-- we und en ersetzen
	rd:		out std_logic;					-- en ersetzen
	addr:		out std_logic_vector(31 downto 0);		-- Adresse an Ausgang (zum Memory)
	di:		out std_logic_vector(31 downto 0);		-- Dateneingang vom Memory
	do:		in std_logic_vector(31 downto 0)		-- Datenausgang zum Memory
    );
end memoryinterface;

architecture rtl of memoryinterface is

	signal vload : std_logic;
	signal vdata_buffer : vectordata_type := (others => (others => '0'));
	signal inc: std_logic;
	signal res: std_logic;
	signal index: integer range 0 to k-1;
	signal counter : std_logic_vector(31 downto 0) := (others => '0');

	type statetype is (waiting, rd_state_0, rd_state_1, rd_state_2, wr_state_0, wr_state_1, 
	       wr_state_2, vrd_state_0, vrd_state_1, vrd_state_2, vwr_state_0, vwr_state_1, vwr_state_2, vdone1, vdone2);
	signal state : statetype := waiting;
	signal nextstate : statetype := waiting;

begin
	-- derive index from counter
	index <= conv_integer(counter) when (counter < k) else 0;
	-- counter
	process
	begin
		wait until clk='1' and clk'event;
		counter <= counter;
		if reset = '1' then
			counter <= (others => '0');
		else
			if res = '1' then
				counter <= (others => '0');
			else
				if inc = '1' then
					counter <= counter + '1';
				end if;
			end if;
		end if;
	end process;

	-- Vector_buffer
	process 
	begin
		wait until clk='1' and clk'event;
		if vload = '1' then
			vdata_buffer(index) <= do;
		else
			vdata_buffer(index) <= vdata_buffer(index);
		end if;
	end process;

	-- state register
	process
	begin
		wait until clk='1' and clk'event;
		state <= nextstate;
	end process;

	-- state transition
	process (clk, address, access_type, data_in, vdata_in, state, do, index, counter)
	begin
		cpu_ready <= '0';
		nextstate <= waiting;

		res <= '0';
		inc <= '0';

		vload <= '0';

		wr <= '0';
		rd <= '0';
		addr <= (others => '0');
		di <= (others => '0');

		case state is
		-- WAITING STATE
		when waiting =>
			cpu_ready <= '0';

			case access_type is
				when "010" =>   -- scalar read
					nextstate <= rd_state_0;
				when "100" =>   -- scalar write
					nextstate <= wr_state_0;
				when "011" =>   -- vector read
					nextstate <= vrd_state_0; 
				when "101" =>   -- scalar write
					nextstate <= vwr_state_0;
				when others =>  
					nextstate <= waiting;
			end case;

		--READ STATE 0
		when rd_state_0 =>
			if mem_ready = '0' then	-- Warten bis Speicher bereit
				nextstate <= rd_state_1;
			else 
				nextstate <= rd_state_0;
			end if;

		--READ STATE 1	
		when rd_state_1 =>
			rd <= '1';
			addr <= address;
			cpu_ready <= '0';
			if mem_ready = '1' then	-- Warten bis Daten anliegen
				nextstate <= rd_state_2;
			else 
				nextstate <= rd_state_1;
			end if;

		-- READ DONE STATE
		when rd_state_2 =>
			rd <= '1';
			addr <= address;
			cpu_ready <= '1';
			if access_type = "000" then
				nextstate <= waiting;
			else
				nextstate <= rd_state_2;
			end if;

		--WRITE STATE 0
		when wr_state_0 =>
			addr <= address; -- mal gucken obs was hilft...
			if mem_ready = '0' then	-- Warten bis Speicher bereit
				nextstate <= wr_state_1;
			else 
				nextstate <= wr_state_0;
			end if;

		--WRITE STATE 1
		when wr_state_1 =>
			 wr <= '1';
			addr <= address;
			di <= data_in;
			cpu_ready <= '0';
			if mem_ready = '1' then	-- Warten bis Daten gespeichert
				nextstate <= wr_state_2;
			else 
				nextstate <= wr_state_1;
			end if; 

		--WRITE DONE STATE
		when wr_state_2 =>
			addr <= address; --wird weiter angelegt um fehler zu vermeiden
			cpu_ready <= '1';
			if access_type = "000" then
				nextstate <= waiting;
			else
				nextstate <= wr_state_2;
			end if;

		-- VECTOR READ STATES
		--VECTOR READ STATE 0 
		when vrd_state_0 =>
			if mem_ready = '0' then	-- Warten bis Speicher bereit
				nextstate <= vrd_state_1;
			else 
				nextstate <= vrd_state_0;
			end if;

		when vrd_state_1 =>
			rd <= '1';
			addr <= address + counter;
			if mem_ready = '1' then  -- Warten bis Daten anliegen
				nextstate <= vrd_state_2;
			else
				nextstate <= vrd_state_1;
			end if;

		when vrd_state_2 =>
			rd <= '1';
			addr <= address + counter;

			inc <= '1';
			vload <= '1';

			if counter = k-1 then  -- �berpr�fen ob alle Vektoreinheiten beladen
				nextstate <= vdone1;
			else
				nextstate <= vrd_state_0;
			end if;

		-- VECTOR WRITE STATES
		--VECTOR WRITE STATE 0 
		when vwr_state_0 =>
			addr <= address + counter;
			if mem_ready = '0' then	-- Warten bis Speicher bereit
				nextstate <= vwr_state_1;
			else 
				nextstate <= vwr_state_0;
			end if;

		when vwr_state_1 =>
			wr <= '1';
			addr <= address + counter;
			di <= vdata_in(index);
			if mem_ready = '1' then  -- Warten bis Daten anliegen
				nextstate <= vwr_state_2;
			else
				nextstate <= vwr_state_1;
			end if;

		when vwr_state_2 =>
			addr <= address + counter; --Adressbus weiter anliegen lassen um fehler zu vermeiden
			inc <= '1';

			if counter = k-1 then
				nextstate <= vdone1;
			else
				nextstate <= vwr_state_0;
			end if;

		-- VECTOR DONE STATE
		when vdone1 =>
			res <= '1';
			nextstate <= vdone2;

		when vdone2 =>
			cpu_ready <= '1';

			if access_type = "000" then
				nextstate <= waiting;
			else
				nextstate <= vdone2;
			end if;
		end case;
	end process;

-- connect outputs
	data_out <= do;
	vdata_out <= vdata_buffer;
end;
