package hicosim.components.backend.processor;

import hicosim.components.backend.processor.Processor.ScalarRegister;
import hicosim.components.backend.processor.enums.Instruction;
import hicosim.components.backend.processor.enums.ScalarType;
import hicosim.components.backend.processor.enums.VecWord;
import hicosim.components.backend.processor.enums.VectorType;

/**
 * Helper for the processor to decode.
 * @author  Michael Wager, Luca Calchera, 16th May 2009
 */
public class Decoder {

	/*
	 * Die Instruction-Arrays:
	 * Die Indexe der BefehlsArrays entsprechen den dekodierten Bits
	 * in den 32 Bit HiCoVec Befehlen.
	 * BSP:
	 * 	LOAD == 1000 = 8 also an 8.Stelle im arrInstrLOAD_STORE 
	 *  ist dann das LOAD Befehl enum.
	 * 
	 */
	private static Instruction [] arrInstrLOAD_STORE = new Instruction[12];
	private static Instruction [] arrInstrALUCommand = new Instruction[16];
	private static Instruction [] arrInstrFLAGCommand = new Instruction[11];
	private static Instruction [] arrInstrJMP_NOPCommand = new Instruction[16];
	
	//Arrays der Skalar-/Vektor-Befehlstypen
	private static VectorType [] arrVectorType = new VectorType[4];
	private static ScalarType [] arrScalarType = new ScalarType[4];
	
	//Array der Vektor Woerter enums 
	private static VecWord []     arrDecipherWord = new VecWord[4];
	
	//Array um fuer das dekodierte Bit ein SkalarRegister enum zu bekommen
	private static ScalarRegister[] arrScalarRegister=new ScalarRegister[4];
	
	/**
	 * Konstruktor - setup the arrays.
	 */
	public Decoder() {
		setUp();
	}
	
	private void setUp() {
		arrInstrLOAD_STORE[8] = Instruction.LOAD;
		arrInstrLOAD_STORE[10] = Instruction.STORE;
		arrInstrLOAD_STORE[9] = Instruction.VLOAD;
		arrInstrLOAD_STORE[11] = Instruction.VSTORE;
		
		arrInstrALUCommand[0] = Instruction.ADD;
		arrInstrALUCommand[1] = Instruction.ADC;
		arrInstrALUCommand[2] = Instruction.SUB;
		arrInstrALUCommand[3] = Instruction.SBC;
		arrInstrALUCommand[4] = Instruction.INC;
		arrInstrALUCommand[6] = Instruction.DEC;
		arrInstrALUCommand[8] = Instruction.AND;
		arrInstrALUCommand[9] = Instruction.OR;
		arrInstrALUCommand[10] = Instruction.XOR;
		arrInstrALUCommand[11] = Instruction.MUL;
		arrInstrALUCommand[12] = Instruction.LSL;
		arrInstrALUCommand[13] = Instruction.ROL;
		arrInstrALUCommand[14] = Instruction.LSR;
		arrInstrALUCommand[15] = Instruction.ROR;
		
		arrInstrFLAGCommand[4] = Instruction.CLC;
		arrInstrFLAGCommand[5] = Instruction.SEC;
		arrInstrFLAGCommand[8] = Instruction.CLZ;
		arrInstrFLAGCommand[10] = Instruction.SEZ;
	
		arrInstrJMP_NOPCommand[8]=Instruction.JMP;
		arrInstrJMP_NOPCommand[12]=Instruction.JNC;
		arrInstrJMP_NOPCommand[13]=Instruction.JC;
		arrInstrJMP_NOPCommand[14]=Instruction.JNZ;
		arrInstrJMP_NOPCommand[15]=Instruction.JZ;
		
		
		arrVectorType[1] = VectorType.VALU;
		arrVectorType[3] = VectorType.VSHUF;
		
		arrScalarType[0] = ScalarType.JMP_NOP;
		arrScalarType[1] = ScalarType.ALU;
		arrScalarType[2] = ScalarType.LOAD_STORE;
		arrScalarType[3] = ScalarType.FLAG;

		arrDecipherWord[0] = VecWord.BYTE;
		arrDecipherWord[1] = VecWord.WORD;
		arrDecipherWord[2] = VecWord.DOUBLEWORD;
		arrDecipherWord[3] = VecWord.QUADWORD;

		arrScalarRegister[1] = ScalarRegister.A;
		arrScalarRegister[2] = ScalarRegister.X;
		arrScalarRegister[3] = ScalarRegister.Y;
	}
	
	/**
	 * This method is used to return a ScalarRegister enum depending on the given intValue.
	 * @param register - int
	 * @return ScalarRegister
	 */
	public ScalarRegister decipherScalarRegister(int register) {
		try {
			return arrScalarRegister[register];
		}
		catch (Exception e) {
			return null;
		}
	}

	/**
	 * This method is used to return a ScalarType enum depending on the given intValue.
	 * @param type - int
	 * @return ScalarType
	 */
	public ScalarType decipherType(int type) {
		try {
			return arrScalarType[type];
		}
		catch (Exception e) {
			return null;
		}
	}
	
	/**
	 * This method is used to return a VecWord enum depending on the given intValue.
	 * @param word - int
	 * @return VecWord
	 */
	public VecWord decipherWord(int word) {
		try {
			return arrDecipherWord[word];
		}
		catch (Exception e) {
			return null;
		}
	}
	
	/**
	 * This method is used to return a VectorType enum depending on the given intValue.
	 * @param type - int
	 * @return VectorType
	 */
	public VectorType decipherVectorType(int type) {
		try {
			return arrVectorType[type];
		}
		catch (Exception e) {
			return null;
		}
	}

	/**
	 * This method is used to return a JMP_NOP-Instruction enum depending on the given intValue.
	 * @param command - int
	 * @return Instruction
	 */
	public Instruction decipherJMP_NOPCommand(int command) {
		try {
			return arrInstrJMP_NOPCommand[command];
		}
		catch (Exception e) {
			return null;
		}
	}

	/**
	 * This method is used to return a FLAG-Instruction enum depending on the given intValue.
	 * @param command - int
	 * @return Instruction
	 */
	public Instruction decipherFLAGCommand(int command) {
		try {
			return arrInstrFLAGCommand[command];
		}
		catch (Exception e) {
			return null;
		}
	}

	/**
	 * This method is used to return a LOAD-STORE-Instruction enum depending on the given intValue.
	 * @param command - int
	 * @return Instruction
	 */
	public Instruction decipherLOAD_STORECommand(int command) {
		try {
			return arrInstrLOAD_STORE[command];
		}
		catch (Exception e) {
			return null;
		}
	}
	/**
	 * This method is used to return a SCALAR-ALU-Instruction enum depending on the given intValue.
	 * @param command - int
	 * @return Instruction
	 */
	public Instruction decipherALUCommand(int command) {
		try {
			return arrInstrALUCommand[command];
		}
		catch (Exception e) {
			return null;
		}
	}
}