package hicosim.components.frontend.tablemodels;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Vector;

import javax.swing.table.AbstractTableModel;

/**
 * This class extends the super class AbstractTableModel and provides methods to
 * hide and show rows and columns and also to add and remove rows and columns.
 *
 * @author Daniel Obermueller 28th May 2009
 *
 */
public class CustomAbstractTableModel extends AbstractTableModel {

	private static final long serialVersionUID = 1L;
	private Vector<Vector<Object>> data;
	private ArrayList<Integer> visibleRows;
	private ArrayList<Integer> visibleColumns;
	private int rows;
	private int columns;

	/**
	 * @param Vector
	 *            <Vector<Object>> data
	 */
	public CustomAbstractTableModel(Vector<Vector<Object>> data) {
		this.data = data;
		visibleRows = new ArrayList<Integer>();
		visibleColumns = new ArrayList<Integer>();
		/* Set the correct size of the JTable. */
		for (int i = 0; i < data.size(); i++) {
			visibleRows.add(i);
			rows++;
		}
		for (int i = 0; i < data.get(0).size(); i++) {
			visibleColumns.add(i);
			columns++;
		}
	}

	public CustomAbstractTableModel() {
		visibleRows = new ArrayList<Integer>();
		visibleColumns = new ArrayList<Integer>();
		data = new Vector<Vector<Object>>();
	}

	/**
	 * This method sets the size of the JTable.
	 *
	 * @author Daniel Obermueller & Sebastian Minning 14th May 2009
	 * @param int
	 *            rows
	 * @param int
	 *            columns
	 */
	public void setJTableSize(int rows, int columns) {
		columns -= this.columns;
		rows -= this.rows;
		/* If the amount of columns stay the same. */
		if (columns == 0)
			;
		/* Else if columns need to be added. */
		else if (columns > 0) {
			for (int i = 0; i < columns; i++)
				addColumn();
		} /* ELse if columns need to removed. */
		else {
			columns *= -1;
			for (int i = 0; i < columns; i++)
				removeLastColumn();
		}
		/* If the amount of rows stay the same. */
		if (rows == 0)
			;
		/* Else if rows need to be added. */
		else if (rows > 0) {
			for (int i = 0; i < rows; i++)
				addRow();
		}
		/* ELse if rows need to removed. */
		else {
			rows *= -1;
			for (int i = 0; i < rows; i++)
				removeLastRow();
		}
	}

	/**
	 * This method returns the amount of visible rows as an Integer-Object. This
	 * method had to be overwritten in order to implement the hide and show
	 * methods.
	 *
	 * @author Daniel Obermueller 28th May 2009
	 * @return int
	 */
	public int getRowCount() {
		return visibleRows.size();
	}

	/**
	 * This method returns the amount of visible columns as an Integer-Object.
	 * This method had to be overwritten in order to implement the hide and show
	 * methods.
	 *
	 * @author Daniel Obermueller 28th May 2009
	 * @return int
	 */
	public int getColumnCount() {
		return visibleColumns.size();
	}

	/**
	 * This method returns the amount of rows as an Integer-Object.
	 *
	 * @author Daniel Obermueller 28th May 2009
	 * @return int
	 */
	public int getAmountOfRows() {
		return rows;
	}

	/**
	 * This method returns the amount of columns as an Integer-Object.
	 *
	 * @author Daniel Obermueller 28th May 2009
	 * @return int
	 */
	public int getAmountOfColumns() {
		return columns;
	}

	/**
	 * This method returns a value, which is an Object, located at the passed
	 * indexes int "rowIndex" and int "columnIndex".
	 *
	 * @author Daniel Obermueller 28th May 2009
	 * @param int
	 *            rowIndex
	 * @param int
	 *            columnIndex
	 * @return Object
	 */
	public Object getValueAt(int rowIndex, int columnIndex) {
		return data.get(visibleRows.get(new Integer(rowIndex))).get(
				visibleColumns.get(new Integer(columnIndex)));
	}

	/**
	 * This method is used to set the data vector of the JTable, which is using
	 * this class.
	 *
	 * @author Daniel Obermueller 28th May 2009
	 * @param Vector
	 *            <Vector<Object>> data
	 */
	public void setData(Vector<Vector<Object>> data) {
		this.data = data;
	}

	/**
	 * This method adds a column at the end of the JTable.
	 *
	 * @author Daniel Obermueller 28th May 2009
	 */
	public void addColumn() {
		/* Add an empty column to each row of the JTable. */
		for (Vector<Object> row : data)
			row.add("");
		visibleColumns.add(columns);
		super.fireTableStructureChanged();
		columns++;
	}

	/**
	 * This method removes the last column of the JTable.
	 *
	 * @author Daniel Obermueller 28th May 2009
	 */
	public void removeLastColumn() {
		/* If there are columns, remove the last column of the JTable. */
		if (columns != 0) {
			columns--;
			/* If there are rows, remove the last column of each row. */
			if (rows != 0) {
				for (Vector<Object> row : data)
					row.removeElementAt(columns);
			}
		}
		visibleColumns.remove(new Integer(columns));
		super.fireTableStructureChanged();
	}

	/**
	 * This method adds a row at the end of the JTable.
	 *
	 * @author Daniel Obermueller 28th May 2009
	 */
	public void addRow() {
		/* Create a new row. */
		Vector<Object> row = new Vector<Object>();
		/*
		 * If there are columns, expand the new row with the correct amount of
		 * empty columns.
		 */
		if (columns != 0) {
			for (int i = 0; i < columns; i++) {
				/* If columns are hidden, show them. */
				if (!visibleColumns.contains(new Integer(i)))
					visibleColumns.add(i);
				row.add("");
			}
		}
		visibleRows.add(rows);
		data.add(row);
		rows++;
		super.fireTableStructureChanged();
	}

	/**
	 * This method removes the last row of the JTable.
	 *
	 * @author Daniel Obermueller 28th May 2009
	 */
	public void removeLastRow() {
		/* If there are rows, remove the last row of the JTable. */
		if (rows != 0) {
			rows--;
			data.removeElementAt(rows);
			visibleRows.remove(new Integer(rows));
			super.fireTableStructureChanged();
		} else
			return;
	}

	/**
	 * This method sets the content of the passed row, by using the Vector<Object>
	 * "rowData".
	 *
	 * @author Daniel Obermueller 28th May 2009
	 * @param Vector
	 *            <Object> rowData
	 * @param int
	 *            row
	 */
	public void setRow(Vector<Object> rowData, int row) {
		data.set(row, rowData);
		super.fireTableDataChanged();
	}

	/**
	 * This method sets the content of the passed cell, specified by the two
	 * Integer-Objects "rowIndex" and "columnIndex", using the Object value.
	 *
	 * @author Daniel Obermueller 28th May 2009
	 * @param Object
	 *            value
	 * @param int
	 *            rowIndex
	 * @param int
	 *            columnIndex
	 */
	public void setValueAt(Object value, int rowIndex, int columnIndex) {
		data.get(rowIndex).set(columnIndex, value);
		super.fireTableDataChanged();
	}



	/**
	 * This method hides a row, which is specified by the passed Integer-Object
	 * "row".
	 *
	 * @author Daniel Obermueller 28th May 2009
	 * @param int
	 *            row
	 */
	public void hideRow(int row) {
		/*
		 * If the passed row was hidden, invoke the method
		 * "fireTableStructureChanged" in order to notify the JTable about the
		 * change and achieve a repaint of it.
		 */
		if (visibleRows.remove(new Integer(row)))
			super.fireTableStructureChanged();
	}

	/**
	 * This method shows a row, which is specified by the passed Integer-Object
	 * "row".
	 *
	 * @author Daniel Obermueller 28th May 2009
	 * @param int
	 *            row
	 */
	public void showRow(int row) {
		/*
		 * If the passed row is already visible or if there are no rows or if
		 * the passed row is out of range, do nothing, else show it.
		 */
		if (visibleRows.contains(new Integer(row)) || rows == 0
				|| row >= rows || row < 0)
			return;
		visibleRows.add(row);
		Collections.sort(visibleRows);
		super.fireTableStructureChanged();
	}

	/**
	 * This method hides a column, which is specified by the passed
	 * Integer-Object "column".
	 *
	 * @author Daniel Obermueller 28th May 2009
	 * @param int
	 *            column
	 */
	public void hideColumn(int column) {
		/*
		 * If the passed column was hidden, invoke the method
		 * "fireTableStructureChanged" in order to notify the JTable about the
		 * change and achieve a repaint of it.
		 */
		if (visibleColumns.remove(new Integer(column)))
			super.fireTableStructureChanged();
	}

	/**
	 * This method shows a column, which is specified by the passed
	 * Integer-Object "column".
	 *
	 * @author Daniel Obermueller 28th May 2009
	 * @param int
	 *            column
	 */
	public void showColumn(int column) {
		/*
		 * If the passed column is already visible or if there are no columns or
		 * if the passed column is out of range, do nothing, else show it.
		 */
		if (visibleColumns.contains(new Integer(column)) || rows == 0
				|| column >= columns || columns == 0 || column < 0)
			return;
		visibleColumns.add(column);
		Collections.sort(visibleColumns);
		super.fireTableStructureChanged();
	}
}
