package hicosim.control.utils;

import java.util.*;

/**
 * Class Statistics collects information about excecuted processor-commands.
 * The class also provides methods to manage observers and to notify them about
 * used commands.
 * 
 * @author Weber
 *
 */
public class Statistics implements IStatistic {

	private ArrayList<IStatisticObserver> observers;

	private HashMap<String, Long> statsMap = new HashMap<String, Long>();
	private HashMap<String, Long> statisticTypes = new HashMap<String, Long>();
	
	/* Default-Groups for statistic-commands */
	private String[] types = {
			"SKALARALU",
			"LOADS",
			"STORES",
			"JUMPS",
			"FLAG",
			"MOVES",
			"VECTORLOADS",
			"VECTORSTORES",
			"VECTORALU",
			"OTHER"
	};
	
	
	
	
	

	public Statistics() {
		observers = new ArrayList<IStatisticObserver>();
		
		/* Insert defined groups in hashmap statisticTypes */
		for(int i = 0; i < types.length; i++){
			statisticTypes.put(types[i], new Long(0));
		}
		
	}

	/**
	 * registerObservers - registers new observers who want to be informed on changes
	 * of comands
	 * 
	 * @param IStatisticObserver newObserver 
	 */
	public void registerObserver(IStatisticObserver newObserver) {
		observers.add(newObserver);
	}

	/**
	 * deleteObservers - deletes observer
	 * 
	 * @param IStatisticObserver oldObserver 
	 */
	public void deleteObserver(IStatisticObserver oldObserver) {
		int i = observers.indexOf(oldObserver);
		if (i >= 0) {
			observers.remove(i);
		}
	}

	/**
	 *  updateStatistics - This method is invoked by the class processor on every
	 *  execution of a comand. The method manages the statistics (comand and type
	 *  of comand) in 2 hashmaps. The method also calls the update-method of all its
	 *  observers.
	 *  
	 *  @param String comand - e.g. Add, Vmov etc.
	 */
	public void updateStatistics(String comand) {
		/* Map wird durchsucht, um festzustellen ob Befehl schon einmal ausgefuehrt wurde.
		 * Wenn ja -> Befehl um 1 erhoehen
		 * Wenn nein -> Befehl der Map hinzufuegen mit Wert 1
		 */
		if (statsMap.containsKey(comand)) {
			long anz = statsMap.get(comand);
			statsMap.put(comand, ++anz);
		} else {
			statsMap.put(comand, new Long(1));
		}
		
		/* festlegen des Typs und erhohen des dazugehoerigen Wertes */
		String type = determineType(comand);
		
		long anzTyp = statisticTypes.get(type);
	
		statisticTypes.put(type, ++anzTyp);

		for (int i = 0; i < observers.size(); i++) {
			observers.get(i).updateStatsObserver(comand, type);
		}
	}
	
	/**
	 * reset - Resets the whole statistics and also calls the method reset on
	 * all observers
	 */
	public void reset(){
		if(statsMap.size() > 0){
			this.statsMap.clear();
		}

		/* Statistik-Typen auf 0 zuruecksetzen */
		/* Insert defined groups in hashmap statisticTypes */
		for(int i = 0; i < types.length; i++){
			statisticTypes.put(types[i], new Long(0));
		}
		
		/* execute reset-comand on observers */
		for (int i = 0; i < observers.size(); i++) {
			observers.get(i).resetStats();
		}
	}
	
	/**
	 * getStatistics() - return the collected statistics in an ArrayList which
	 * consists of 2 hasmaps - (1) for the comands and (2) for the type of
	 * comands (e.g. "SKALARALU etc.")
	 * 
	 * @return ArrayList stats
	 */
	public ArrayList<HashMap<String, Long>> getStatistics(){
		ArrayList<HashMap<String, Long>> stats = new ArrayList<HashMap<String, Long>>();
		stats.add(statsMap);
		stats.add(statisticTypes);
		return stats;
	}
	
	/**
	 * getTypes() - return the default comand-types (e.g. "SKALARALU")
	 * 
	 * @return String[] types
	 */
	public String[] getTypes(){
		return this.types;
	}
	
	
	/**
	 * determines type of command and returns it as a string
	 * new commands which are not implemented in the statistics yet
	 * are automatically grouped to "OTHER"
	 * 
	 * @param String command - e.g. "Add"
	 * @return String type - e.g. "SKALARALU"
	 */
	private String determineType(String comand){
	
		/* Load */
		if(comand == "Load"){
			return "LOADS";
		}
		
		/* Store */
		if(comand == "Store"){
			return "STORES";
		}
		
		/* Vload */
		if(comand == "Vload"){
			return "VECTORLOADS";
		}
		
		/* Vstore */
		if(comand == "Vstore"){
			return "VECTORSTORES";
		}
		
		/* SKALARALU */
		if(comand == "Add"){
			return "SKALARALU";
		}		
		if(comand == "Sub"){
			return "SKALARALU";
		}		
		if(comand == "Adc"){
			return "SKALARALU";
		}		
		if(comand == "Sbc"){
			return "SKALARALU";
		}		
		if(comand == "Inc"){
			return "SKALARALU";
		}		
		if(comand == "Dec"){
			return "SKALARALU";
		}		
		if(comand == "And"){
			return "SKALARALU";
		}		
		if(comand == "Or"){
			return "SKALARALU";
		}		
		if(comand == "Xor"){
			return "SKALARALU";
		}		
		if(comand == "Mul"){
			return "SKALARALU";
		}		
		if(comand == "Lsl"){
			return "SKALARALU";
		}		
		if(comand == "Lsr"){
			return "SKALARALU";
		}		
		if(comand == "Rol"){
			return "SKALARALU";
		}		
		if(comand == "Ror"){
			return "SKALARALU";
		}		
		
		/* VektorALU */
		if(comand == "Vadd"){
			return "VECTORALU";
		}	
		if(comand == "Vand"){
			return "VECTORALU";
		}	
		if(comand == "VSub"){
			return "VECTORALU";
		}		
		if(comand == "Vand"){
			return "VECTORALU";
		}		
		if(comand == "Vor"){
			return "VECTORALU";
		}		
		if(comand == "Vxor"){
			return "VECTORALU";
		}		
		if(comand == "Vadd"){
			return "Vmul";
		}		
		if(comand == "Vlsl"){
			return "VECTORALU";
		}		
		if(comand == "Vlsr"){
			return "VECTORALU";
		}
		if(comand == "Vmul"){
			return "VECTORALU";
		}
		
		/* Jump-Befehle */
		if(comand == "Jmp"){
			return "JUMPS";
		}
		if(comand == "Jal"){
			return "JUMPS";
		}		
		if(comand == "Jnz"){
			return "JUMPS";
		}		
		if(comand == "Jz"){
			return "JUMPS";
		}		
		if(comand == "Jnc"){
			return "JUMPS";
		}		
		if(comand == "Jc"){
			return "JUMPS";
		}
		
		/* FLAG-Befehle */
		if(comand == "Clc"){
			return "FLAG";
		}		
		if(comand == "Sec"){
			return "FLAG";
		}		
		if(comand == "Clz"){
			return "FLAG";
		}		
		if(comand == "Sez"){
			return "FLAG";
		}		
		
		/* MOV-Befehle */
		if(comand == "Mov"){
			return "MOVES";
		}		
		if(comand == "Mova"){
			return "MOVES";
		}		
		if(comand == "Vmov"){
			return "MOVES";
		}		
		if(comand == "Vmor"){
			return "MOVES";
		}		
		if(comand == "Vmol"){
			return "MOVES";
		}				
		
		/* Default return */
		return "OTHER";
				
	}

	/**
	 * String representation, e.g. for console output.
	 * Commands and types are sorted by frequency.
	 * 
	 * @return String
	 */
	public String toString(){
		/* Vorbereiten der Ausgabe: Sortieren der Maps */
		LinkedHashMap<String, Long> orderedTypes = sortStatsMaps(statisticTypes);
		LinkedHashMap<String, Long> orderedComands = sortStatsMaps(statsMap);
						
		int totalInstr = 0;
				
		StringBuffer s = new StringBuffer();
		s.append("\n----------------------------\n");
		s.append("HiCoSim Statistics:\n");
		s.append("----------------------------\n");
		s.append("\n");

		for(Map.Entry<String, Long> e: orderedTypes.entrySet()){
			s.append(e.getKey() + ": " + e.getValue());
			s.append("\n");
			
			for(Map.Entry<String, Long> f: orderedComands.entrySet()){
				if(determineType(f.getKey()).equalsIgnoreCase(e.getKey())){
					s.append("  ");
					s.append(f.getKey() + ": " + f.getValue());
					s.append("\n");
					
					totalInstr += f.getValue();
				}
				
			}
		}
				
		s.append("\n");
		s.append("Total executed instructions: " + totalInstr);
		
		s.trimToSize();
		
		return s.toString();
		
	}
	
	/**
	 * private method to sort a given HashMap<String, Integer> by its values
	 * @param defMap (HashMap<String, Integer>);
	 * @return newMap (LinkedHashMap<String, Integer>)
	 */
	private LinkedHashMap<String, Long>  sortStatsMaps(HashMap<String, Long> defMap){
		HashMap<String, Long> oldMap = new HashMap<String, Long>();
		
		/* Map kopieren */
		for(Map.Entry<String, Long> e : defMap.entrySet()){
			oldMap.put(e.getKey(), e.getValue());
		}
		
		List<String> mapKeys = new ArrayList<String>(oldMap.keySet());
		List <Long> mapValues = new ArrayList<Long>(oldMap.values());
		
		Collections.sort(mapValues);
		Collections.sort(mapKeys);
		Collections.reverse(mapValues);
		
		LinkedHashMap<String, Long> newMap = new LinkedHashMap<String, Long>();
		
		Iterator<Long> valueIt = mapValues.iterator();
		while(valueIt.hasNext()){
			Object val = valueIt.next();
			Iterator<String> keyIt = mapKeys.iterator();
			
			while(keyIt.hasNext()){
				Object key = keyIt.next();
				String vgl1 =oldMap.get(key).toString();
				String vgl2 = val.toString();
				
				if(vgl1.equals(vgl2)){
					oldMap.remove(key);
					mapKeys.remove(key);
					newMap.put((String)key, (Long)val);
					break;
				}
			}
		}
		mapKeys = null;
		mapValues = null;
		return newMap;
	}
}
